# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""

DataWig HPO
Implements hyperparameter optimisation for datawig

"""

import itertools
import time
import random
import os

import numpy as np
import pandas as pd

from datetime import datetime
from pandas.api.types import is_numeric_dtype
from sklearn.metrics import mean_squared_error, f1_score, precision_score, accuracy_score, recall_score

from .column_encoders import BowEncoder, CategoricalEncoder, NumericalEncoder, ColumnEncoder, TfIdfEncoder
from .mxnet_input_symbols import BowFeaturizer, NumericalFeaturizer, Featurizer, EmbeddingFeaturizer
from .utils import logger, get_context, random_split, rand_string, flatten_dict, merge_dicts


class _HPO:
    """
    Implements systematic hyperparameter optimisation for datawig

    Example usage:

    imputer = SimpleImputer(input_columns, output_column)
    hps = dict( ... )  # specify hyperparameter choices
    hpo = HPO(impter, hps)
    results = hpo.tune

    """

    def __init__(self):

        """
        Init method also defines default hyperparameter choices, global and for each input column type.

        :param imputer: SimpleImputer instance with input_colums and output_column and output_path specified.

        """

        self.hps = None
        self.results = pd.DataFrame()
        self.output_path = None

        # Define default hyperparameter choices for each column type (string, categorical, numeric)
        default_hps = dict()
        default_hps['global'] = {}
        default_hps['global']['learning_rate'] = [3e-4]
        default_hps['global']['weight_decay'] = [0, 1e-7]
        default_hps['global']['num_epochs'] = [100]
        default_hps['global']['patience'] = [5]
        default_hps['global']['batch_size'] = [16]
        default_hps['global']['final_fc_hidden_units'] = [[], [100]]
        default_hps['global']['concat_columns'] = [True, False]

        default_hps['string'] = {}
        default_hps['string']['ngram_range'] = {}
        default_hps['string']['max_tokens'] = [2 ** 15]
        default_hps['string']['tokens'] = [['words']]
        default_hps['string']['ngram_range']['words'] = [(1, 3)]
        default_hps['string']['ngram_range']['chars'] = [(1, 5)]

        default_hps['categorical'] = {}
        default_hps['categorical']['max_tokens'] = [2 ** 12, 2 ** 15]
        default_hps['categorical']['embed_dim'] = [10]

        default_hps['numeric'] = {}
        default_hps['numeric']['normalize'] = [True]
        default_hps['numeric']['numeric_latent_dim'] = [10, 50]
        default_hps['numeric']['numeric_hidden_layers'] = [1, 2]

        # parameters for a single column of concatenated strings
        default_hps['concat'] = default_hps['string'].copy()

        self.default_hps = default_hps

    def __preprocess_hps(self,
                         train_df: pd.DataFrame,
                         simple_imputer,
                         num_evals) -> pd.DataFrame:
        """
        Generates list of all possible combinations of hyperparameter from the nested hp dictionary.
        Requires the data to check whether the relevant columns are present and have the appropriate type.

        :param train_df: training data as dataframe
        :param simple_imputer: Parent instance of SimpleImputer
        :param num_evals is the maximum number of hpo configurations to consider.

        :return: Data frame where each row is a hyperparameter configuration and each column is a parameter.
                    Column names have the form colum:parameter, e.g. title:max_tokens or global:learning rate.
        """

        # create empty dict if global hps not passed
        if 'global' not in self.hps.keys():
            self.hps['global'] = {}

        # create empty dict if parameters for concatenated column not passed
        if 'concat' not in self.hps.keys():
            self.hps['concat'] = {}

        # add type to column dictionaries if it was not specified, does not support categorical types
        for column_name in simple_imputer.input_columns:
            if column_name not in self.hps.keys():
                self.hps[column_name] = {}
            if 'type' not in self.hps[column_name].keys():
                if is_numeric_dtype(train_df[column_name]):
                    self.hps[column_name]['type'] = ['numeric']
                else:
                    self.hps[column_name]['type'] = ['string']

        # check that all passed parameter are valid hyperparameters
        assert all([key in self.default_hps['global'].keys() for key in self.hps['global'].keys()])
        for key, val in self.hps.items():
            if 'type' in val.keys():
                assert all([key in list(self.default_hps[val['type'][0]].keys()) + ['type'] for key, _ in val.items()])

        # augment provided global self.hps with default global hps so that cartesian products are full parameter sets.
        # self.hps['global'] = {**self.default_hps['global'], **self.hps['global']}
        self.hps['global'] = merge_dicts(self.default_hps['global'], self.hps['global'])
        self.hps['concat'] = merge_dicts(self.default_hps['concat'], self.hps['concat'])

        # augment provided self.hps with default self.hps, iterating over every input column
        for column_name in simple_imputer.input_columns:
            # add dictionary for input columns where no hyperparamters have been passed
            if column_name not in self.hps.keys():
                self.hps[column_name] = {}
                if column_name in simple_imputer.numeric_columns:
                    self.hps[column_name]['type'] = ['numeric']
                elif column_name in simple_imputer.string_columns:
                    self.hps[column_name]['type'] = ['string']
                else:
                    logger.warn('Input type of column ' + str(column_name) + ' not determined.')
            # join all column specific hp dictionaries with type-specific default values
            # self.hps[column_name] = {**self.default_hps[self.hps[column_name]['type'][0]], **self.hps[column_name]}
            self.hps[column_name] = merge_dicts(self.default_hps[self.hps[column_name]['type'][0]],
                                                self.hps[column_name])

        # flatten nested dictionary structures and combine to data frame with all possible hp configurations
        hp_df_from_dict = lambda dict: pd.DataFrame(list(itertools.product(*dict.values())), columns=dict.keys())

        from datawig.utils import random_cartesian_product
        flat_dict = flatten_dict(self.hps)
        hp_df = pd.DataFrame(random_cartesian_product([val for key, val in flat_dict.items()], num=num_evals),
                             columns=flat_dict.keys())

        return hp_df

    def __fit_hp(self,
                 train_df: pd.DataFrame,
                 test_df: pd.DataFrame,
                 hp: pd.Series,
                 simple_imputer,
                 name: str,
                 user_defined_scores: list = None) -> pd.core.series.Series:
        """

        Method initialises the model, performs fitting and returns the desired metrics.


        :param train_df: training data as dataframe
        :param test_df: test data as dataframe; if not provided, a ratio of test_split of the
                          training data are used as test data
        :param hp: pd.Series with hyperparameter configuration
        :param simple_imputer: SimpleImputer instance from which to inherit column names etc.
        :param name to identify the current setting of hps.
        :param user_defined_scores: list with entries (Callable, str), where callable is a function
                          accepting arguments (true, predicted, confidence). True is an array with the true labels,
                          predicted with the predicted labels and confidence is an array with the confidence score for
                          each prediction.
                          Default metrics are:
                          f1_weighted, f1_micro, f1_macro, f1_weighted_train
                          recall_weighted, recall_weighted_train, precision_weighted, precision_weighted_train,
                          coverage_at_90, coverage_at_90_train, empirical_precision_at_90,
                          ece_pre_calibration (ece: expected calibration error), ece_post_calibration, time [min].
                          A user defined function could look as follows:

                          def my_function(true, predicted, confidence):
                               return (true[confidence > .75] == predicted[confidence > .75]).mean()

                          uds = (my_function, 'empirical_precision_above_75')

        :return: Series with hpo parameters and results.

        """

        from . import Imputer  # needs to be imported here to avoid circular dependency

        if not name:
            name = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        data_encoders = []
        data_featurizers = []

        if hp['global:concat_columns'] is False:

            # mark unused parameter
            for key, val in hp.items():
                if 'concat:' in key:
                    hp[key] = 'n.a.'

            # define column encoders and featurisers for each input column
            for input_column in simple_imputer.input_columns:

                # extract parameters for the current input column, take everything after the first colon
                col_parms = {':'.join(key.split(':')[1:]): val for key, val in hp.items() if input_column in key}

                # define all input columns
                if col_parms['type'] == 'string':
                    # iterate over multiple embeddings (chars + strings for the same column)
                    for token in col_parms['tokens']:
                        # call kw. args. with: **{key: item for key, item in col_parms.items() if not key == 'type'})]
                        data_encoders += [TfIdfEncoder(input_columns=[input_column],
                                                       output_column=input_column + '_' + token,
                                                       tokens=token,
                                                       ngram_range=col_parms['ngram_range:'+token],
                                                       max_tokens=col_parms['max_tokens'])]
                        data_featurizers += [BowFeaturizer(field_name=input_column + '_' + token,
                                                           max_tokens=col_parms['max_tokens'])]

                elif col_parms['type'] == 'categorical':
                    data_encoders += [CategoricalEncoder(input_columns=[input_column],
                                                         output_column=input_column + '_' + col_parms['type'],
                                                         max_tokens=col_parms['max_tokens'])]
                    data_featurizers += [EmbeddingFeaturizer(field_name=input_column + '_' + col_parms['type'],
                                                             max_tokens=col_parms['max_tokens'],
                                                             embed_dim=col_parms['embed_dim'])]

                elif col_parms['type'] == 'numeric':
                    data_encoders += [NumericalEncoder(input_columns=[input_column],
                                                       output_column=input_column + '_' + col_parms['type'],
                                                       normalize=col_parms['normalize'])]
                    data_featurizers += [NumericalFeaturizer(field_name=input_column + '_' + col_parms['type'],
                                                             numeric_latent_dim=col_parms['numeric_latent_dim'],
                                                             numeric_hidden_layers=col_parms['numeric_hidden_layers'])]
                else:
                    logger.warn('Found unknown column type. Canidates are string, categorical, numeric.')

        # Concatenate all columns
        else:
            # cast all columns to string for concatenation
            train_df = train_df.astype(str)
            test_df = test_df.astype(str)

            col_parms = {':'.join(key.split(':')[1:]): val for key, val in hp.items() if 'concat' in key}
            for token in col_parms['tokens']:
                data_encoders += [TfIdfEncoder(input_columns=simple_imputer.input_columns,
                                               output_column='-'.join(simple_imputer.input_columns) + '_' + token,
                                               tokens=token,
                                               ngram_range=col_parms['ngram_range:' + token],
                                               max_tokens=col_parms['max_tokens'])]
                data_featurizers += [BowFeaturizer(field_name='-'.join(simple_imputer.input_columns) + '_' + token,
                                                   max_tokens=col_parms['max_tokens'])]
                # mark unused parameter
                for key, val in hp.items():
                    if not ('global:' in key or 'concat:' in key):
                        hp[key] = 'n.a.'

        # Define separate encoder and featurizer for each column
        # Define output column. Associated parameters are not tuned.
        if is_numeric_dtype(train_df[simple_imputer.output_column]):
            label_column = [NumericalEncoder(simple_imputer.output_column)]
            logger.info("Assuming numeric output column: {}".format(simple_imputer.output_column))
        else:
            label_column = [CategoricalEncoder(simple_imputer.output_column)]
            logger.info("Assuming categorical output column: {}".format(simple_imputer.output_column))

        global_parms = {key.split(':')[1]: val for key, val in hp.iteritems() if 'global' in key}

        hp_time = time.time()

        hp_imputer = Imputer(data_encoders=data_encoders,
                             data_featurizers=data_featurizers,
                             label_encoders=label_column,
                             output_path=self.output_path + name)

        hp_imputer.fit(train_df=train_df,
                       test_df=test_df,
                       ctx=get_context(),
                       learning_rate=global_parms['learning_rate'],
                       num_epochs=global_parms['num_epochs'],
                       patience=global_parms['patience'],
                       test_split=.1,
                       weight_decay=global_parms['weight_decay'],
                       batch_size=global_parms['batch_size'],
                       final_fc_hidden_units=global_parms['final_fc_hidden_units'],
                       calibrate=True)

        # add suitable metrics to hp series
        imputed = hp_imputer.predict(test_df)
        true = imputed[simple_imputer.output_column]
        predicted = imputed[simple_imputer.output_column + '_imputed']

        imputed_train = hp_imputer.predict(train_df.sample(min(train_df.shape[0], int(1e4))))
        true_train = imputed_train[simple_imputer.output_column]
        predicted_train = imputed_train[simple_imputer.output_column + '_imputed']

        if is_numeric_dtype(train_df[simple_imputer.output_column]):
            hp['mse'] = mean_squared_error(true, predicted)
            hp['mse_train'] = mean_squared_error(true_train, predicted_train)
            confidence = float('nan')
        else:
            confidence = imputed[simple_imputer.output_column + '_imputed_proba']
            confidence_train = imputed_train[simple_imputer.output_column + '_imputed_proba']
            hp['f1_micro'] = f1_score(true, predicted, average='micro')
            hp['f1_macro'] = f1_score(true, predicted, average='macro')
            hp['f1_weighted'] = f1_score(true, predicted, average='weighted')
            hp['f1_weighted_train'] = f1_score(true_train, predicted_train, average='weighted')
            hp['precision_weighted'] = f1_score(true, predicted, average='weighted')
            hp['precision_weighted_train'] = f1_score(true_train, predicted_train, average='weighted')
            hp['recall_weighted'] = recall_score(true, predicted, average='weighted')
            hp['recall_weighted_train'] = recall_score(true_train, predicted_train, average='weighted')
            hp['coverage_at_90'] = (confidence > .9).mean()
            hp['coverage_at_90_train'] = (confidence_train > .9).mean()
            hp['empirical_precision_at_90'] = (predicted[confidence > .9] == true[confidence > .9]).mean()
            hp['ece_pre_calibration'] = hp_imputer.calibration_info['ece_post']
            hp['ece_post_calibration'] = hp_imputer.calibration_info['ece_post']
            hp['time [min]'] = (time.time() - hp_time)/60

        for uds in user_defined_scores:
            hp[uds[1]] = uds[0](true=true, predicted=predicted, confidence=confidence)

        hp_imputer.save()

        return hp

    def tune(self,
             train_df: pd.DataFrame,
             test_df: pd.DataFrame = None,
             hps: dict = None,
             strategy: str = 'random',
             num_evals: int = 10,
             max_running_hours: float = 96,
             user_defined_scores: list = None,
             hpo_run_name: str = None,
             simple_imputer = None):

        """
        Do grid search or random search for hyperparameter configurations. This method can not tune tfidf vs hashing
        vectorization but uses tfidf. Also parameters of the output column encoder are not tuned.
        
        :param train_df: training data as dataframe
        :param test_df: test data as dataframe; if not provided, a ratio of test_split of the
                          training data are used as test data
        :param hps: nested dictionary where hps[global][parameter_name] is list of parameters. Similarly,
                          hps[column_name][parameter_name] is a list of parameter values for each input column.
                          Further, hps[column_name]['type'] is in ['numeric', 'categorical', 'string'] and is
                          inferred if not provided. See init method of HPO for further details.

        :param strategy: 'random' for random search or 'grid' for exhaustive search
        :param num_evals: number of evaluations for random search
        :param max_running_hours: Time before the hpo run is terminated in hours
        :param user_defined_scores: list with entries (Callable, str), where callable is a function
                          accepting **kwargs true, predicted, confidence. Allows custom scoring functions.
        :param hpo_run_name: Optional string identifier for this run.
                          Allows to sequentially run hpo jobs and keep previous iterations
        :param simple_imputer: SimpleImputer instance from which to inherit column names etc.

        :return: None
        """

        self.output_path = simple_imputer.output_path

        if user_defined_scores is None:
            user_defined_scores = []

        if hpo_run_name is None:
            hpo_run_name = ""

        self.hps = hps
        simple_imputer.check_data_types(train_df)  # infer data types, saved self.string_columns, self.numeric_columns

        # train/test split if no test data given
        if test_df is None:
            train_df, test_df = random_split(train_df, [.8, .2])

        # process_hp_configurations(hps) and return random configurations
        hps_flat = self.__preprocess_hps(train_df, simple_imputer, num_evals)

        logger.info("Training starts for " + str(hps_flat.shape[0]) + "hyperparameter configurations.")

        # iterate over hp configurations and fit models. This loop could be parallelized
        start_time = time.time()
        elapsed_time = 0

        for hp_idx, (_, hp) in enumerate(hps_flat.iterrows()):
            if elapsed_time > max_running_hours:
                logger.info('Finishing hpo because max running time was reached.')
                break

            # if concat_columns True, set all n.a. hps to n.a. TODO
            logger.info("Fitting hpo iteration " + str(hp_idx) + " with parameters\n\t" +
                        '\n\t'.join([str(i) + ': ' + str(j) for i, j in hp.items()]))
            name = hpo_run_name + str(hp_idx)

            # add results to hp
            hp = self.__fit_hp(train_df,
                               test_df,
                               hp,
                               simple_imputer,
                               name,
                               user_defined_scores)

            # append output to results data frame
            self.results = pd.concat([self.results, hp.to_frame(name).transpose()])

            # save to file in every iteration
            if not os.path.exists(simple_imputer.output_path):
                os.makedirs(simple_imputer.output_path)
            self.results.to_csv(os.path.join(simple_imputer.output_path, "hpo_results.csv"))

            logger.info('Finished hpo iteration ' + str(hp_idx))
            elapsed_time = (time.time() - start_time)/3600

        logger.info('Assigning model with highest weighted precision to SimpleImputer object and copying artifacts.')
        simple_imputer.load_hpo_model()  # assign best model to simple_imputer.imputer and write artifacts.



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Column Encoders:
used for translating values of a table into numerical representation such that Featurizers can
operate on them
"""

import os
import random
from abc import ABCMeta, abstractmethod
from functools import partial
from typing import Any, Dict, Iterable, List

import mxnet as mx
import numpy as np
import pandas as pd
from sklearn.feature_extraction.text import HashingVectorizer, TfidfVectorizer
from sklearn.preprocessing import StandardScaler

from .utils import logger

random.seed(0)
np.random.seed(42)


class NotFittedError(BaseException):
    """

    Error thrown when unfitted encoder is used

    """


class ColumnEncoder():
    """

    Abstract super class of column encoders.
    Transforms value representation of columns (e.g. strings) into numerical representations to be
    fed into MxNet.

    Options for ColumnEncoders are:

        SequentialEncoder:  for sequences of symbols (e.g. characters or words),
        BowEncoder: bag-of-word representation, as sparse vectors
        CategoricalEncoder: for categorical variables
        NumericalEncoder: for numerical values

    :param input_columns: List[str] with column names to be used as input for this ColumnEncoder
    :param output_column: Name of output field, used as field name in downstream MxNet iterator
    :param output_dim: dimensionality of encoded column values (1 for categorical, vocabulary size
                    for sequential and BoW)

    """
    __metaclass__ = ABCMeta

    def __init__(self,
                 input_columns: List[str],
                 output_column=None,
                 output_dim=1):

        if not isinstance(input_columns, list):
            input_columns = [input_columns]

        for col in input_columns:
            if not isinstance(col, str):
                raise ValueError("ColumnEncoder.input_columns must be str type, was {}".format(type(col)))

        if output_column is None:
            output_column = "-".join(input_columns)
            logstr = "No output column name provided for ColumnEncoder " \
                     "using {}".format(output_column)
            logger.info(logstr)

        self.input_columns = input_columns
        self.output_column = output_column
        self.output_dim = int(output_dim)


    @abstractmethod
    def transform(self, data_frame: pd.DataFrame) -> np.array:
        """

        Transforms values in one or more columns of DataFrame into a numpy array that can be fed
        into a Featurizer

        :param data_frame:
        :return: List of integers

        """
        pass

    @abstractmethod
    def fit(self, data_frame: pd.DataFrame):
        """

        Fits a ColumnEncoder if needed (i.e. vocabulary/alphabet)

        :param data_frame: pandas DataFrame
        :return:

        """
        return self

    @abstractmethod
    def is_fitted(self):
        """

        Checks if ColumnEncoder (still) needs to be fitted to data

        :return: True if the column encoder does not require fitting (anymore or at all)

        """
        pass

    @abstractmethod
    def decode(self, col: pd.Series) -> pd.Series:
        """

        Decodes a pandas Series of token indices

        :param col: pandas Series of token indices
        :return: pandas Series of tokens

        """
        pass


class CategoricalEncoder(ColumnEncoder):
    """

    Transforms categorical variable from string representation into number

    :param input_columns: List[str] with column names to be used as input for this ColumnEncoder
    :param output_column: Name of output field, used as field name in downstream MxNet iterator
    :param token_to_idx: token to index mapping,
            0 is reserved for missing tokens, 1 ... max_tokens for most to least frequent tokens
    :param max_tokens: maximum number of tokens

    """

    def __init__(self,
                 input_columns: Any,
                 output_column: str = None,
                 token_to_idx: Dict[str, int] = None,
                 max_tokens: int = int(1e4)) -> None:

        ColumnEncoder.__init__(self, input_columns, output_column, 1)

        if len(self.input_columns) != 1:
            raise ValueError("CategoricalEncoder can only encode single columns, got {}: {}".format(
                len(self.input_columns), ", ".join(self.input_columns)))

        self.max_tokens = int(max_tokens)
        self.token_to_idx = token_to_idx
        self.idx_to_token = None

    @staticmethod
    def transform_func_categorical(col: pd.Series,
                                   token_to_idx: Dict[str, int],
                                   missing_token_idx: int) -> Any:
        """

        Transforms categorical values into their indices

        :param col: pandas Series with categorical values
        :param token_to_idx: Dict[str, int] with mapping from token to token index
        :param missing_token_idx: index for missing symbol
        :return:

        """
        return [token_to_idx.get(v, missing_token_idx) for v in col]

    def is_fitted(self):
        """

        Checks if ColumnEncoder (still) needs to be fitted to data

        :return: True if the column encoder does not require fitting (anymore or at all)

        """
        return self.token_to_idx is not None

    def transform(self, data_frame: pd.DataFrame) -> np.array:
        """

        Transforms string column of pandas dataframe into categoricals

        :param data_frame: pandas data frame
        :return: numpy array (rows by 1)

        """

        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        if not self.token_to_idx:
            raise NotFittedError("CategoricalEncoder needs token to index mapping")

        if not self.idx_to_token:
            self.idx_to_token = {idx: token for token, idx in self.token_to_idx.items()}

        func = partial(self.transform_func_categorical, token_to_idx=self.token_to_idx,
                       missing_token_idx=0)

        logger.info("CategoricalEncoder encoding {} rows with \
                    {} tokens from column {} to \
                    column {}".format(len(data_frame), len(self.token_to_idx), self.input_columns[0], self.output_column))

        return data_frame[self.input_columns].apply(func).values.astype(np.float32)

    def fit(self, data_frame: pd.DataFrame):
        """

        Fits a CategoricalEncoder by extracting the value histogram of a column and capping it at
        max_tokens. Issues warning if less than 100 values were observed.

        :param data_frame: pandas data frame

        """

        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        value_histogram = data_frame[self.input_columns[0]].replace('',
                                                                    np.nan).dropna().value_counts()

        self.max_tokens = int(min(len(value_histogram), self.max_tokens))

        self.token_to_idx = {token: idx + 1 for idx, token in
                             enumerate(value_histogram.index[:self.max_tokens])}
        self.idx_to_token = {idx: token for token, idx in self.token_to_idx.items()}
        logger.info("{} most often encountered discrete values: \
                     {}".format(self.max_tokens,value_histogram.index.values[:self.max_tokens]))

        for label in value_histogram.index[:self.max_tokens]:
            if value_histogram[label] < 100:
                logger.warning("CategoricalEncoder for column {} \
                               found only {} occurrences of value {}".format(self.input_columns[0], value_histogram[label], label))

        return self

    def decode_token(self, token_idx: int) -> str:
        """

        Decodes a token index into a token

        :param token_idx: token index
        :return: token

        """
        return self.idx_to_token.get(token_idx, "MISSING")

    def decode(self, col: pd.Series) -> pd.Series:
        """

        Decodes a pandas Series of token indices

        :param col: pandas Series of token indices
        :return: pandas Series of tokens

        """
        return col.map(self.idx_to_token).fillna("MISSING")


class SequentialEncoder(ColumnEncoder):
    """

    Transforms sequence of characters into sequence of numbers

    :param input_columns: List[str] with column names to be used as input for this ColumnEncoder
    :param output_column: Name of output field, used as field name in downstream MxNet iterator
    :param token_to_idx: token to index mapping
           0 is reserved for missing tokens, 1 ... max_tokens-1 for most to least frequent tokens
    :param max_tokens: maximum number of tokens
    :param seq_len: length of sequence, shorter sequences get padded to, longer sequences
                    truncated at seq_len symbols

    """

    def __init__(self,
                 input_columns: Any,
                 output_column: str = None,
                 token_to_idx: Dict[str, int] = None,
                 max_tokens: int = int(1e3),
                 seq_len: int = 500) -> None:

        ColumnEncoder.__init__(self, input_columns, output_column, seq_len)

        if len(self.input_columns) != 1:
            raise ValueError("SequentialEncoder can only encode single columns, got {}: {}".format(
                len(self.input_columns), ", ".join(self.input_columns)))

        self.token_to_idx = token_to_idx
        self.idx_to_token = None
        self.max_tokens = int(max_tokens)

    @staticmethod
    def transform_func_seq_single(string: str,
                                  token_to_idx: Dict[str, int],
                                  seq_len: int,
                                  missing_token_idx: int) -> List[int]:
        """

        Transforms a single string into a sequence of token ids

        :param string: a sequence of symbols as string
        :param token_to_idx: Dict[str, int] with mapping from token to token index
        :param seq_len: length of sequence
        :param missing_token_idx: index for missing symbol
        :return: List[int] with transformed values

        """
        if isinstance(string, str):
            rep = [token_to_idx.get(tok, missing_token_idx) for tok in string[:seq_len]]
        else:
            rep = []
        pad = [missing_token_idx] * (seq_len - len(rep))
        return rep + pad

    def is_fitted(self) -> bool:
        """

        Checks if ColumnEncoder (still) needs to be fitted to data

        :return: True if the column encoder does not require fitting (anymore or at all)

        """
        return self.token_to_idx is not None

    def fit(self, data_frame: pd.DataFrame):
        """

        Fits a SequentialEncoder by extracting the character value histogram of a column and
        capping it at max_tokens

        :param data_frame: pandas data frame

        """

        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        logger.info("Fitting SequentialEncoder on %s rows", str(len(data_frame)))

        # get the relevant column and concatenate all strings
        flattened = pd.Series(
            list(data_frame[self.input_columns].astype(str).replace(' ', '').fillna(
                '').values.sum()))
        # compute character histograms, take most frequent ones
        char_hist = flattened.value_counts().sort_values(ascending=False)[:self.max_tokens]

        logstr = "Characters encoded for {}: {}".format(self.input_columns[0],
                                                        "".join(sorted(char_hist.index.values)))
        logger.info(logstr)

        self.max_tokens = int(min(len(char_hist), self.max_tokens))
        self.token_to_idx = {token: idx + 1 for token, idx in
                             zip(char_hist.index, range(self.max_tokens))}
        self.idx_to_token = {idx: char for char, idx in self.token_to_idx.items()}

        return self

    def transform(self, data_frame: pd.DataFrame) -> np.array:
        """

        Transforms column of pandas dataframe into sequence of tokens

        :param data_frame: pandas DataFrame
        :return: numpy array (rows by seq_len)

        """

        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        if not self.token_to_idx:
            raise NotFittedError("SequentialEncoder needs token to index mapping")

        logstr = "Applying SequentialEncoder on {} rows with {} tokens and seq_len {} " \
                 "from column {} to column {}".format(len(data_frame), len(self.token_to_idx),
                                                      self.output_dim, self.input_columns[0],
                                                      self.output_column)
        logger.info(logstr)

        func = partial(self.transform_func_seq_single, token_to_idx=self.token_to_idx,
                       seq_len=self.output_dim,
                       missing_token_idx=0)

        return np.vstack(data_frame[self.input_columns[0]].apply(func).values).astype(np.float32)

    def decode_seq(self, token_index_sequence: Iterable[int]) -> str:
        """

        Decodes a sequence of token indices into a string

        :param token_index_sequence: an iterable of token indices
        :return: str the decoded string

        """
        return "".join([self.idx_to_token.get(token_idx, "") for token_idx in token_index_sequence])

    def decode(self, col: pd.Series) -> pd.Series:
        """
        Decodes a pandas Series of token indices

        :param col: pandas Series of token index iterables
        :return: pd.Series of strings

        """
        return col.apply(self.decode_seq).fillna("MISSING")


class TfIdfEncoder(ColumnEncoder):
    """

    TfIdf bag of word encoder for text data, using sklearn's TfidfVectorizer

    :param input_columns: List[str] with column names to be used as input for this ColumnEncoder
    :param output_column: Name of output field, used as field name in downstream MxNet iterator
    :param max_tokens: Number of feature buckets (dimensionality of sparse ngram vector). default 2**18
    :param tokens: How to tokenize the input data, supports 'words' and 'chars'.
    :param ngram_range: length of ngrams to use as features
    :param prefixed_concatenation: whether or not to prefix values with column name before concat

    """

    def __init__(self,
                 input_columns: Any,
                 output_column: str = None,
                 max_tokens: int = 2 ** 18,
                 tokens: str = 'chars',
                 ngram_range: tuple = None,
                 prefixed_concatenation: bool = True) -> None:

        ColumnEncoder.__init__(self, input_columns, output_column, int(max_tokens))

        if ngram_range is None:
            ngram_range = (1, 3) if tokens == 'words' else (1, 5)

        if tokens == 'words':
            self.vectorizer = TfidfVectorizer(max_features=self.output_dim, ngram_range=ngram_range)
        elif tokens == 'chars':
            self.vectorizer = TfidfVectorizer(max_features=self.output_dim, ngram_range=ngram_range,
                                              analyzer="char")
        else:
            logger.info(
                "BowEncoder attribute tokens has to be 'words' or 'chars', defaulting to 'chars'")
            self.vectorizer = TfidfVectorizer(max_features=self.output_dim,
                                              ngram_range=ngram_range, analyzer="char")

        self.prefixed_concatenation = prefixed_concatenation

        self.idx_to_token = None

        self.__fitted = False

    def __preprocess_input(self, data_frame: pd.DataFrame) -> pd.Series:
        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        if len(self.input_columns) == 1:
            tmp_col = data_frame[self.input_columns[0]]
        else:
            tmp_col = pd.Series(index=data_frame.index, data='')
            for col in self.input_columns:
                if self.prefixed_concatenation:
                    tmp_col += col + " " + data_frame[col].fillna("") + " "
                else:
                    tmp_col += data_frame[col].fillna("") + " "
            logstr = "Applying TfIdf BoW Encoding to columns {} {} prefix into column {}".format(
                self.input_columns, "with" if self.prefixed_concatenation else "without",
                self.output_column)
            logger.info(logstr)

        return tmp_col

    def fit(self, data_frame: pd.DataFrame):
        """
        :param data_frame:
        :return:

        """
        self.vectorizer.fit(self.__preprocess_input(data_frame))
        self.idx_to_token = {idx: token for token, idx in self.vectorizer.vocabulary_.items()}
        self.__fitted = True
        return self

    def is_fitted(self) -> bool:
        """
        :param self:
        :return: True if the encoder is fitted

        """
        return self.__fitted

    def transform(self, data_frame: pd.DataFrame) -> np.array:
        """
        Transforms one or more string columns into Bag-of-words vectors.

        :param data_frame: pandas DataFrame with text columns
        :return: numpy array (rows by max_features)

        """
        return self.vectorizer.transform(self.__preprocess_input(data_frame)).astype(np.float32)

    def decode(self, col: pd.Series) -> pd.Series:
        """

        Given a series of indices, decode it to input tokens

        :param col:
        :return: pd.Series of tokens

        """
        return col.map(lambda index: self.idx_to_token[index])


class BowEncoder(ColumnEncoder):
    """

    Bag-of-Words encoder for text data, using sklearn's HashingVectorizer

    :param input_columns: List[str] with column names to be used as input for this ColumnEncoder
    :param output_column: Name of output field, used as field name in downstream MxNet iterator
    :param max_tokens: Number of hash buckets (dimensionality of sparse ngram vector). default 2**18
    :param tokens: How to tokenize the input data, supports 'words' and 'chars'.
    :param ngram_range: length of ngrams to use as features
    :param prefixed_concatenation: whether or not to prefix values with column name before concat

    """

    def __init__(self,
                 input_columns: Any,
                 output_column: str = None,
                 max_tokens: int = 2 ** 18,
                 tokens: str = 'chars',
                 ngram_range: tuple = None,
                 prefixed_concatenation: bool = True) -> None:

        if ngram_range is None:
            ngram_range = (1, 3) if tokens == 'words' else (1, 5)

        ColumnEncoder.__init__(self, input_columns, output_column, int(max_tokens))

        if tokens == 'words':
            self.vectorizer = HashingVectorizer(n_features=self.output_dim, ngram_range=ngram_range)
        elif tokens == 'chars':
            self.vectorizer = HashingVectorizer(n_features=self.output_dim, ngram_range=ngram_range,
                                                analyzer="char")
        else:
            logger.info(
                "BowEncoder attribute tokens has to be 'words' or 'chars', defaulting to 'chars'")
            self.vectorizer = HashingVectorizer(n_features=self.output_dim, ngram_range=ngram_range,
                                                analyzer="char")

        self.prefixed_concatenation = prefixed_concatenation

    def fit(self, data_frame: pd.DataFrame):
        """

        Does nothing, HashingVectorizers do not need to be fit.

        :param data_frame:
        :return:

        """
        logger.info("BowEncoder is stateless and doesn't need to be fit")
        return self

    def is_fitted(self) -> bool:
        """

        Returns true if the column encoder does not require fitting (anymore or at all)

        :param self:
        :return: True if the encoder is fitted

        """
        return True

    def transform(self, data_frame: pd.DataFrame) -> np.array:
        """

        Transforms one or more string columns into Bag-of-words vectors, hashed into a max_features
        dimensional feature space. Nans and missing values will be replaced by zero vectors.

        :param data_frame: pandas DataFrame with text columns
        :return: numpy array (rows by max_features)

        """
        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        if len(self.input_columns) == 1:
            tmp_col = data_frame[self.input_columns[0]]
        else:
            tmp_col = pd.Series(index=data_frame.index, data='')
            for col in self.input_columns:
                if self.prefixed_concatenation:
                    tmp_col += col + " " + data_frame[col].fillna("") + " "
                else:
                    tmp_col += data_frame[col].fillna("") + " "
            logstr = "Applying Hashing BoW Encoding to columns {} {} prefix into column {}".format(
                self.input_columns, "with" if self.prefixed_concatenation else "without",
                self.output_column)
            logger.info(logstr)

        return self.vectorizer.transform(tmp_col).astype(np.float32)

    def decode(self, col: pd.Series) -> pd.Series:
        """

        Raises NotImplementedError, hashed bag-of-words cannot be decoded due to hash collisions

        :param token_index_sequence:

        :return:

        """
        raise NotImplementedError


class NumericalEncoder(ColumnEncoder):
    """

    Numerical encoder, concatenates columns in field_names into one vector
    fills nans with the mean of a column

    :param input_columns: List[str] with column names to be used as input for this ColumnEncoder
    :param output_column: Name of output field, used as field name in downstream MxNet iterator
    :param normalize: whether to normalize by the standard deviation or not, default True

    """

    def __init__(self,
                 input_columns: Any,
                 output_column: str = None,
                 normalize=True) -> None:

        ColumnEncoder.__init__(self, input_columns, output_column, 0)

        self.output_dim = len(self.input_columns)
        self.normalize = normalize
        self.scaler = None

    def is_fitted(self):
        """

        Returns true if the column encoder does not require fitting (anymore or at all)

        :param self:
        :return: True if the encoder is fitted

        """
        fitted = True

        if self.normalize:
            fitted = self.scaler is not None

        return fitted

    def fit(self, data_frame: pd.DataFrame):
        """

        Does nothing or fits the normalizer, if normalization is specified

        :param data_frame: DataFrame with numerical columns specified when
                            instantiating NumericalEncoder

        """
        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        mean = data_frame[self.input_columns].mean()
        data_frame[self.input_columns] = data_frame[self.input_columns].fillna(mean)
        self.scaler = StandardScaler().fit(data_frame[self.input_columns].values)

        return self

    def transform(self, data_frame: pd.DataFrame) -> np.array:
        """

        Concatenates the numerical columns specified when instantiating the NumericalEncoder
        Normalizes features if specified in the NumericalEncoder

        :param data_frame: DataFrame with numerical columns specified in NumericalEncoder
        :return: np.array with numerical features (rows by number of numerical columns)

        """
        if not isinstance(data_frame, pd.core.frame.DataFrame):
            raise ValueError("Only pandas data frames are supported")

        if self.scaler is None:
            self.scaler = StandardScaler().fit(data_frame[self.input_columns].values)

        mean = pd.Series(dict(zip(self.input_columns,self.scaler.mean_)))
        data_frame[self.input_columns] = data_frame[self.input_columns].fillna(mean)

        logger.info("Concatenating numeric columns %s into %s",
                    self.input_columns,
                    self.output_column)

        if self.normalize:
            logger.info("Normalizing with StandardScaler")
            encoded = self.scaler.transform(data_frame[self.input_columns].values).astype(
                np.float32)
        else:
            encoded = data_frame[self.input_columns].values.astype(np.float32)
        return encoded

    def decode(self, col: pd.Series) -> pd.Series:
        """

        Undoes the normalization, scales by scale and adds the mean

        :param col: pandas Series (normalized)
        :return: pandas Series (unnormalized)

        """

        if self.normalize:
            decoded = (col * self.scaler.scale_) + self.scaler.mean_
        else:
            decoded = col

        return decoded



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""

Methods for model calibration via temperature scaling,
applied as post-processing step.

"""


import numpy as np
from scipy.optimize import minimize
from .utils import softmax


def compute_ece(scores: np.ndarray,
                labels: np.ndarray,
                lbda: float = 1,
                step: float = .05) -> float:
    """

    :param scores: Probabilities or logit scores of dimension samples x classes.
    :param labels: True labels as corresponding column indices
    :param lbda: scaling parameter, lbda = 1 amounts to no rescaling.
    :param step: histogram bin width. Can be tuned.

    :return:

    """

    # compute probabilities applying tempered softmax
    probas = calibrate(scores, lbda)

    # probabilities for the most likely labels
    top_probas = np.max(probas, axis=1)

    # top prediction is independent of temperature scaling or logits/probas
    predictions = np.argmax(probas, 1)

    bin_means = np.arange(step / 2, 1, step)

    ece = 0

    # iterate over bins and compare confidence and precision
    for bin_lower, bin_upper in [(mean - step / 2, mean + step / 2) for mean in bin_means]:

        # select bin entries
        bin_mask = (top_probas >= bin_lower) & (top_probas < bin_upper)
        if np.any(bin_mask):
            in_bin_confidence = np.mean(top_probas[bin_mask])
            in_bin_precision = np.mean(labels[bin_mask] == predictions[bin_mask])
            ece += np.abs(in_bin_confidence - in_bin_precision) * np.mean(bin_mask)

    return ece


def ece_loss(lbda: float, *args) -> float:
    """

    :param lbda: Scaling parameter
    wrapper around compute_ece() to be called during optimisation

    """

    scores, labels = args

    return compute_ece(scores, labels, lbda=lbda)


def logits_from_probas(probas: np.ndarray, force: bool=False) -> np.ndarray:
    """
    Returns logits for a vector of class probabilities. This is not a unique transformation.
    If the input is not a probability, no transformation is made by default.

    :param probas: Probabilities of dimension samples x classes.
    :param force: True forces rescaling, False only rescales if the values look like probabilities.
    """

    # check whether rows of input are probabilities
    if (force is True) or (np.all(np.sum(probas, 1) - 1 < 1e-12) and np.all(probas <= 1)):
        return np.log(probas)
    else:
        return probas


def probas_from_logits(scores: np.ndarray, lbda: float=1, force: bool=False) -> np.ndarray:
    """
    Returns probabilitiess for a vector of class logits.
    If the input is a probability, no transformation is made.

    :param scores: Logits of dimension samples x classes.
    :param lbda: parameter for temperature scaling
    :param force: True forces rescaling, False only rescales if the values don't look like probabilities.
    """

    # check whether rows of input are probabilities
    if np.all(np.sum(scores, 1) - 1 < 1e-2) and np.all(scores <= 1) and force is False:
        return scores
    else:
        return np.array([softmax(lbda * row) for row in scores])


def calibrate(scores: np.ndarray, lbda: float) -> np.ndarray:
    """
    Apply temperature scaling

    :param scores: Probabilities of dimension samples x classes. Do not pass logits.
    :param lbda: Parameter for temperature scaling.
    :return: Calibrated array of probabilities of dimensions samples x classes.
    """

    logits = logits_from_probas(scores, force=True)

    return np.array([softmax(lbda * row) for row in logits])


def reliability(scores: np.ndarray, labels: np.ndarray, step: float=.05) -> tuple:
    """
    Compute tuples for reliability plots.

    :param scores: Probabilities or logits of dimension samples x classes.
    :param labels: True labels as corresponding column indices
    :param step: histogram bin width. Can be tuned.
    :return: tuple containing mean of bins and the precision in each bin.
    """

    # transform scores to probabilities if applicable
    probas = probas_from_logits(scores)

    # probabilities for the most likely labels
    top_probas = np.max(probas, axis=1)

    predictions = np.argmax(probas, 1)

    bin_means = np.arange(step / 2, 1, step)

    in_bin_precisions = np.zeros(len(bin_means))

    for i, (bin_lower, bin_upper) in enumerate([(mean - step / 2, mean + step / 2) for mean in bin_means]):

        bin_mask = (top_probas >= bin_lower) & (top_probas < bin_upper)
        if np.any(bin_mask):
            in_bin_precisions[i] = np.mean(labels[bin_mask] == predictions[bin_mask])

    return bin_means, in_bin_precisions


def fit_temperature(scores: np.ndarray, labels: np.ndarray) -> float:
    """
    Find temperature scaling parameter through optimisting the expected calibration error.

    :param scores: Probabilities or logits of dimension samples x classes.
    :param labels: True labels as corresponding column indices
    :return: temperature scaling parameter lbda
    """

    probas = probas_from_logits(scores)

    res = minimize(ece_loss, 1, method='SLSQP', tol=1e-6, args=(probas, labels),
                   options={'maxiter': 10000}, bounds=((1e-10, 100),))

    assert res['success'] == True

    return res['x'][0]





# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""

DataWig ColumnEncoder tests

"""

import os

import numpy as np
import pandas as pd
import pytest

from datawig import column_encoders

df = pd.DataFrame({'features': ['xwcxG pQldP Cel0n 5LaWO 2cjTu',
                                '2cjTu YizDY u1aEa Cel0n SntTK',
                                '2cjTu YizDY u1aEa Cel0n SntTK'],
                   'labels': ['xwcxG', 'SntTK', 'SntTK']})

categorical_encoder = column_encoders.CategoricalEncoder(['labels'], max_tokens=3).fit(df)
sequential_encoder = column_encoders.SequentialEncoder(['features'],
                                                       max_tokens=50, seq_len=3).fit(df)

# CategoricalEncoder Tests
def test_categorical_encoder_unfitted_fail():
    unfitted_categorical_encoder = column_encoders.CategoricalEncoder(["col_1"])
    assert not unfitted_categorical_encoder.is_fitted()
    with pytest.raises(column_encoders.NotFittedError):
        unfitted_categorical_encoder.transform(pd.DataFrame({"col_1": ['a', 'b']}))


def test_fit_categorical_encoder():
    assert categorical_encoder.is_fitted()
    assert categorical_encoder.token_to_idx == {'SntTK': 1, 'xwcxG': 2}
    assert categorical_encoder.idx_to_token == {1: 'SntTK', 2: 'xwcxG'}


def test_categorical_encoder_transform():
    assert categorical_encoder.transform(df).flatten()[0] == 2.


def test_categorical_encoder_transform_missing_token():
    assert categorical_encoder.transform(pd.DataFrame({'labels': ['foobar']})).flatten()[0] == 0


def test_categorical_encoder_max_token():
    categorical_encoder = column_encoders.CategoricalEncoder(['labels'], max_tokens=1e4).fit(df)
    assert categorical_encoder.max_tokens == 2


def test_categorical_encoder_decode_token():
    assert categorical_encoder.decode_token(1) == 'SntTK'


def test_categorical_encoder_decode_missing_token():
    assert categorical_encoder.decode_token(0) == 'MISSING'


def test_categorical_encoder_decode():
    assert categorical_encoder.decode(pd.Series([1])).values[0] == 'SntTK'


def test_categorical_encoder_decode_missing():
    assert categorical_encoder.decode(pd.Series([0])).values[0] == 'MISSING'


def test_categorical_encoder_non_negative_embedding_indices():
    assert all(categorical_encoder.transform(df).flatten() >= 0)


# SequentialEncoder Tests
def test_sequential_encoder_unfitted_fail():
    unfitted_sequential_encoder = column_encoders.SequentialEncoder(["col_1"])
    assert not unfitted_sequential_encoder.is_fitted()
    with pytest.raises(column_encoders.NotFittedError):
        unfitted_sequential_encoder.transform(pd.DataFrame({'brand': ['ab']}))


def test_fit_sequential_encoder():
    sequential_encoder_fewer_tokens = column_encoders.SequentialEncoder(['features'],
                                                                        max_tokens=5,
                                                                        seq_len=3).fit(df)
    assert (set(sequential_encoder_fewer_tokens.token_to_idx.keys()) == {'u', 'a', 'n', 'T', ' '})


def test_sequential_encoder_transform():
    encoded = pd.Series([vec.tolist() for vec in sequential_encoder.transform(df)])
    true_decoded = df['features'].apply(lambda x: x[:sequential_encoder.output_dim])
    assert all(sequential_encoder.decode(encoded) == true_decoded)


def test_sequential_encoder_transform_missing_token():
    assert (sequential_encoder.transform(pd.DataFrame({'features': ['!~']}))[0].tolist() == [0, 0,
                                                                                             0])


def test_sequential_encoder_max_token():
    sequential_encoder_short = column_encoders.SequentialEncoder("features", max_tokens=1e4,
                                                                 seq_len=2)
    sequential_encoder_short.fit(df)
    assert sequential_encoder.is_fitted()
    assert sequential_encoder_short.max_tokens == 32


def test_sequential_encoder_non_negative_embedding_indices():
    assert all(sequential_encoder.transform(df).flatten() >= 0)


def test_bow_encoder():
    bow_encoder = column_encoders.BowEncoder("features", max_tokens=5)
    assert bow_encoder.is_fitted()
    bow = bow_encoder.transform(df)[0].toarray()[0]
    true = np.array([0.615587, -0.3077935, -0.3077935, -0.41039133, 0.51298916])
    assert true == pytest.approx(bow, 1e-4)


def test_bow_encoder_multicol():
    bow_encoder = column_encoders.BowEncoder(["item_name", "product_description"], max_tokens=5)
    data = pd.DataFrame({'item_name': ['bla'], 'product_description': ['fasl']})
    bow = bow_encoder.transform(data)[0].toarray()[0]
    true = np.array([0.27500955, -0.82502865, -0.1833397, 0., -0.45834925])
    assert true == pytest.approx(bow, 1e-4)
    data_strings = ['item_name bla product_description fasl ']
    assert true == pytest.approx(bow_encoder.vectorizer.transform(data_strings).toarray()[0])


def test_categorical_encoder_numeric():
    df = pd.DataFrame({'brand': [1, 2, 3]})
    try:
        column_encoders.CategoricalEncoder("brand").fit(df)
    except TypeError:
        pytest.fail("fitting categorical encoder on integers should not fail")


def test_categorical_encoder_numeric_transform():
    df = pd.DataFrame({'brand': [1, 2, 3, 1, 2, 1, np.nan, None]})
    col_enc = column_encoders.CategoricalEncoder("brand").fit(df)
    assert np.array_equal(col_enc.transform(df), np.array([[1], [2], [3], [1], [2], [1], [0], [0]]))


def test_categorical_encoder_numeric_nan():
    df = pd.DataFrame({'brand': [1, 2, 3, None]})
    try:
        column_encoders.CategoricalEncoder("brand").fit(df)
    except TypeError:
        pytest.fail("fitting categorical encoder on integers with nulls should not fail")

def test_column_encoder_no_list_input_column():
    column_encoder = column_encoders.ColumnEncoder("0")
    assert column_encoder.input_columns == ['0']
    assert column_encoder.output_column == '0'
    with pytest.raises(ValueError):
        column_encoders.ColumnEncoder(0)
    with pytest.raises(ValueError):
        column_encoders.ColumnEncoder([0])


def test_numeric_encoder():
    df = pd.DataFrame({'a': [1, 2, 3, np.nan, None], 'b': [.1, -.1, np.nan, None, 10.5]})
    unfitted_numerical_encoder = column_encoders.NumericalEncoder(["a", 'b'], normalize=False)
    assert unfitted_numerical_encoder.is_fitted()
    fitted_unnormalized_numerical_encoder = unfitted_numerical_encoder.fit(df)
    df_unnormalized = fitted_unnormalized_numerical_encoder.transform(df.copy())

    assert np.array_equal(df_unnormalized, np.array([[1., 0.1],
                                                     [2., -0.1],
                                                     [3., 3.5],
                                                     [2.0, 3.5],
                                                     [2.0, 10.5]], dtype=np.float32))
    df_nans = pd.DataFrame({'a': [None], 'b': [np.nan]})
    df_unnormalized_nans = fitted_unnormalized_numerical_encoder.transform(df_nans.copy())
    assert np.array_equal(df_unnormalized_nans, np.array([[2., 3.5]], dtype=np.float32))

    normalized_numerical_encoder = column_encoders.NumericalEncoder(["a", 'b'], normalize=True)
    assert not normalized_numerical_encoder.is_fitted()
    normalized_numerical_encoder_fitted = normalized_numerical_encoder.fit(df)
    df_normalized = normalized_numerical_encoder_fitted.transform(df)

    assert normalized_numerical_encoder.is_fitted()
    assert np.array_equal(df_normalized, np.array([[-1.58113885, -0.88666826],
                                                   [0., -0.93882525],
                                                   [1.58113885, 0.],
                                                   [0., 0.],
                                                   [0., 1.82549345]], dtype=np.float32))

def test_tfidf_encoder():
    tfidf_encoder = column_encoders.TfIdfEncoder("features", max_tokens=5)
    assert tfidf_encoder.is_fitted() is False
    tfidf_encoder.fit(df)
    bow = tfidf_encoder.transform(df)[0].toarray()[0]
    true = np.array([0.75592893, 0.5669467, 0.18898223, 0.18898223, 0.18898223])
    assert tfidf_encoder.is_fitted() is True
    assert true == pytest.approx(bow, 1e-4)

    decoded_indices = tfidf_encoder.decode(pd.Series([0, 1, 2]))
    assert np.array_equal(decoded_indices.values, np.array([' ', 'c', 'e']))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""

DataWig evaluation functions

"""

import json
import pandas as pd
from sklearn.metrics import precision_score, recall_score, f1_score, accuracy_score, \
    precision_recall_curve, mean_squared_error
from .utils import logger


def evaluate_and_persist_metrics(true_labels_string,
                                 true_labels_int,
                                 predictions,
                                 predictions_proba,
                                 metrics_file=None,
                                 missing_symbol=None,
                                 numerical_labels={},
                                 numerical_predictions={}):
    '''
    Compute and save metrics in metrics_file
    :param true_labels_string: dict with true labels e.g. {'color':["black","white"]}
    :param true_labels_int: dict with true label indices e.g. {'color':[0,1]}
    :param predictions: dict with predicted labels e.g. {'color':["black","white"]}
    :param predictions_proba: dict with class likelihoods e.g. {'color':[0.1,0.9]}
    :param true_labels_int: dict with predicted class index e.g. {'color':[1,0]}
    :param metrics_file: path to *.json file to store metrics
    :param missing_symbol: dict with missing symbol (will be excluded from metrics) e.g.
                        {'color': "MISSING"}
    :param numerical_labels: dict with true numerical outputs
    :param numerical_predictions: dict with predicted numerical outputs
    :return:
    '''

    if missing_symbol is None:
        missing_symbol = {k: "MISSING" for k in true_labels_string.keys()}

    if len(true_labels_string) > 0:
        # concatenate all categorical predictions
        dfs = []
        for att in true_labels_string.keys():
            true_and_predicted = [(true, pred) for true, pred in zip(true_labels_string[att], predictions[att]) if
                                  true != missing_symbol[att]]
            assert len(true_and_predicted) > 0, "No valid ground truth data for label {}".format(att)
            truth, predicted = zip(*true_and_predicted)
            logger.info(
                "Keeping {}/{} not-missing values for evaluation of {}".format(len(truth), len(true_labels_string[att]),
                                                                               att))
            dfs.append(pd.DataFrame({'attribute': att, 'true_value': truth, 'predicted_value': predicted}))

        df = pd.concat(dfs)

        metrics = evaluate_model_outputs(df)

        logger.info("average classification metrics:")
        for label, att_metrics in metrics.items():
            logger.info(
                "label: {} - {}".format(label, list(filter(lambda x: x[0].startswith("avg"), att_metrics.items()))))

        logger.info("weighted average classification metrics:")
        for label, att_metrics in metrics.items():
            logger.info(
                "label: {} - {}".format(label,
                                        list(filter(lambda x: x[0].startswith("weighted"), att_metrics.items()))))

        for att in true_labels_string.keys():
            metrics[att]['precision_recall_curves'] = {}
            for label in range(1, predictions_proba[att].shape[-1] + 1):
                true_labels = (true_labels_int[att] == label).nonzero()[0]
                if len(true_labels) > 0:
                    true_label = true_labels_string[att][true_labels[0]]
                    y_true = (true_labels_int[att] == label) * 1.0
                    y_score = predictions_proba[att][:, label]
                    prec, rec, thresholds = precision_recall_curve(y_true, y_score)
                    metrics[att]['precision_recall_curves'][true_label] = {'precision': prec,
                                                                           'recall': rec,
                                                                           'thresholds': thresholds}
                    threshold_idx = (prec > .95).nonzero()[0][0] - 1
                    logger.info(
                        "Attribute {}, Label: {}\tReaching {} precision / {} recall at threshold {}".format(
                            att, true_label, prec[threshold_idx], rec[threshold_idx], thresholds[threshold_idx]))
    else:
        metrics = {}

    for col_name in numerical_labels.keys():
        metrics[col_name] = 1.0 * mean_squared_error(numerical_labels[col_name], numerical_predictions[col_name])

    if metrics_file:
        import copy
        serialize_metrics = copy.deepcopy(metrics)
        # transform precision_recall_curves to json serializable lists
        for att in true_labels_string.keys():
            for label in metrics[att]['precision_recall_curves'].keys():
                serialize_metrics[att]['precision_recall_curves'][label] = \
                    {
                        'precision': metrics[att]['precision_recall_curves'][label]['precision'].tolist(),
                        'recall': metrics[att]['precision_recall_curves'][label]['recall'].tolist(),
                        'thresholds': metrics[att]['precision_recall_curves'][label]['thresholds'].tolist()
                    }
        logger.info("save metrics in {}".format(metrics_file))
        with open(metrics_file, "w") as fp:
            json.dump(serialize_metrics, fp)

    return metrics


def evaluate_model_outputs(
        df,
        attribute_column_name='attribute',
        true_label_column_name='true_value',
        predicted_label_column_name='predicted_value'):
    '''

    :param df: a dataframe with attribute | true_value | predicted_value
    :return: precision/recall/f1/accuracy for each attribute (class frequency weighted) and for each column-label combination
    '''

    groups = df.groupby(attribute_column_name)

    model_metrics = {}

    for group, group_df in groups:
        true = group_df[true_label_column_name]
        predicted = group_df[predicted_label_column_name]

        model_metrics[group] = evaluate_model_outputs_single_attribute(true, predicted)

    return model_metrics


def evaluate_model_outputs_single_attribute(true, predicted, topMisclassifications=100):
    true = true.astype(str)
    predicted = predicted.astype(str)

    labels = true.value_counts()

    model_metrics = dict()

    model_metrics['class_counts'] = [(a, int(b)) for a, b in labels.iteritems()]

    # computes statistics not weighted by class frequency
    model_metrics['avg_precision'] = precision_score(true, predicted, average='macro')
    model_metrics['avg_recall'] = recall_score(true, predicted, average='macro')
    model_metrics['avg_f1'] = f1_score(true, predicted, average='macro')
    model_metrics['avg_accuracy'] = accuracy_score(true, predicted)

    # computes statistics weighted by class frequency
    model_metrics['weighted_precision'] = precision_score(true, predicted, average='weighted')
    model_metrics['weighted_recall'] = recall_score(true, predicted, average='weighted')
    model_metrics['weighted_f1'] = f1_score(true, predicted, average='weighted')
    # todo sample_weight seems missing
    model_metrics['weighted_accuracy'] = accuracy_score(true, predicted)

    # single class metrics
    model_metrics['class_precision'] = precision_score(true, predicted, average=None,
                                                       labels=labels.index.tolist()).tolist()
    model_metrics['class_recall'] = recall_score(true, predicted, average=None, labels=labels.index.tolist()).tolist()
    model_metrics['class_f1'] = f1_score(true, predicted, average=None, labels=labels.index.tolist()).tolist()
    model_metrics['class_accuracy'] = accuracy_score(true, predicted).tolist()
    model_metrics['num_classes'] = len(model_metrics['class_counts'])
    model_metrics['num_applicable_rows'] = int(sum(count for (class_name, count) in model_metrics['class_counts']))

    true_name = "true"
    pred_name = "pred"
    groups = pd.DataFrame(list(zip(true, predicted)), columns=[true_name, pred_name]).groupby(true_name)
    model_metrics['confusion_matrix'] = []
    for label in labels.index.tolist():
        confusion_matrix_series = groups.get_group(label)[pred_name].value_counts()[:topMisclassifications]
        confusion_matrix = list(zip(confusion_matrix_series.index.tolist(), map(int, confusion_matrix_series.tolist())))
        model_metrics['confusion_matrix'].append((label, confusion_matrix))

    return model_metrics



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import random

import shutil

import mxnet as mx
import numpy as np
import pandas as pd
import pytest

from datawig.utils import rand_string


@pytest.fixture(scope='function')
def reset_random_seed():
    # Setting seed for PRNG(s) for every test in the file
    random.seed(0)
    np.random.seed(0)
    mx.random.seed(0)

    yield

@pytest.fixture
def test_dir():
    file_dir = os.path.dirname(os.path.realpath(__file__))
    joined = os.path.join(file_dir, "resources")
    yield joined
    shutil.rmtree(joined)

@pytest.fixture
def data_frame():

    def _inner_impl(
            feature_col='features',
            label_col='labels',
            n_samples=500,
            word_length=5,
            num_words=100,
            vocab_size=100,
            num_labels=10):

        """
        Generates text features and categorical labels.
        :param feature_col: name of feature column.
        :param label_col: name of label column.
        :param n_samples: how many rows to generate.
        :return: pd.DataFrame with columns = [feature_col, label_col]
        """

        vocab = [rand_string(word_length) for i in range(vocab_size)]
        labels = vocab[:num_labels]
        words = vocab[num_labels:]

        def _sentence_with_label(labels=labels, words=words):
            """
            Generates a random token sequence containing a random label

            :param labels: label set
            :param words: vocabulary of tokens
            :return: blank separated token sequence and label

            """
            label = random.choice(labels)
            tokens = [random.choice(words) for _ in range(num_words)] + [label]
            sentence = " ".join(np.random.permutation(tokens))

            return sentence, label

        sentences, labels = zip(*[_sentence_with_label(labels, words) for _ in range(n_samples)])
        df = pd.DataFrame({feature_col: sentences, label_col: labels})

        return df

    return _inner_impl



import os
import setuptools

ROOT = os.path.dirname(__file__)

with open(os.path.join(ROOT,"requirements","requirements.txt")) as f:
    required = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="datawig",
    version="0.1.6",
    author="datawig-dev",
    author_email="datawig-dev@amazon.com",
    maintainer_email='datawig-dev@amazon.com',
    description="Imputation for tables with missing values",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/awslabs/datawig",
    packages=setuptools.find_packages(),
    install_requires=required,
    license='Apache License 2.0',
    python_requires='>=3',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only'
    ],
)




# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""

DataWig imputer iterator tests

"""
import itertools

import numpy as np
import pandas as pd

from datawig.column_encoders import (BowEncoder, CategoricalEncoder,
                                     SequentialEncoder)
from datawig.iterators import ImputerIterDf

feature_col = "features"
label_col = "labels"
max_tokens = 100
num_labels = 10


def test_iter_next_df(data_frame):
    it = get_new_iterator_df(data_frame())
    _, next_batch = next(it), next(it)
    print("Array : " + str(next_batch.label[0].asnumpy()))
    assert ((next_batch.label[0].asnumpy() == np.array([[9.], [9.]])).all())

def test_iter_df_bow(data_frame):
    df = data_frame()
    it = get_new_iterator_df_bow(df)
    tt = next(it)
    bow = tt.data[0].asnumpy()[0, :]
    true = it.data_columns[0].vectorizer.transform([df.loc[0,'features']]).toarray()[0]
    assert (true - bow).sum() < 1e-5


def test_iter_provide_label_or_data_df(data_frame):
    it = get_new_iterator_df(data_frame())
     # pylint: disable=unsubscriptable-object
    assert it.provide_data[0][0] == label_col
    assert it.provide_data[0][1] == (2, 2)
    assert it.provide_label[0][0] == label_col
    assert it.provide_label[0][1] == (2, 1)


def test_iter_index_df(data_frame):
    it = get_new_iterator_df(data_frame())
    idx_it = list(itertools.chain(*[b.index for b in it]))
    idx_true = data_frame().index.tolist()
    assert idx_it == idx_true


def test_iter_decoder_df():
    # draw skewed brands
    brands = [{feature_col: brand} for brand in
              list(map(lambda e: str(int(e)), np.random.exponential(scale=1, size=1000)))]

    brand_df = pd.DataFrame(brands)
    it = ImputerIterDf(brand_df,
                       data_columns=[SequentialEncoder(feature_col, max_tokens=10, seq_len=2)],
                       label_columns=[CategoricalEncoder(feature_col, max_tokens=100)],
                       batch_size=2)
    decoded = it.decode(next(it).label)
    np.testing.assert_array_equal(decoded[0], brand_df[feature_col].head(it.batch_size).values)


def test_iter_padding_offset():
    col = 'brand'
    df = pd.DataFrame(
        [{col: brand} for brand in
         list(map(lambda e: str(int(e)), np.random.exponential(scale=1, size=36)))]
    )
    df_train = df.sample(frac=0.5)
    it = ImputerIterDf(
        df_train,
        data_columns=[BowEncoder(col)],
        label_columns=[CategoricalEncoder(col, max_tokens=5)],
        batch_size=32
    )
    assert it.start_padding_idx == df_train.shape[0]

def get_new_iterator_df_bow(df):
    return ImputerIterDf(df,
                         data_columns=[BowEncoder(feature_col, max_tokens=max_tokens)],
                         label_columns=[CategoricalEncoder(label_col, max_tokens=num_labels)],
                         batch_size=2)


def get_new_iterator_df(df):
    return ImputerIterDf(df,
                         data_columns=[SequentialEncoder(label_col,
                                                         max_tokens=max_tokens,
                                                         seq_len=2)],
                         label_columns=[CategoricalEncoder(label_col, max_tokens=max_tokens)],
                         batch_size=2)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""

DataWig utility functions

"""

import contextlib
import itertools
import logging
import math
import random
import sys
import time
from typing import Any, List, Tuple, Dict
import collections

import mxnet as mx
import numpy as np
import pandas as pd

mx.random.seed(1)
random.seed(1)
np.random.seed(42)

# set global logger variables
log_formatter = logging.Formatter("%(asctime)s [%(levelname)s]  %(message)s")
logger = logging.getLogger()
consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(log_formatter)
logger.addHandler(consoleHandler)
logger.setLevel("INFO")


def flatten_dict(d: Dict,
                 parent_key: str ='',
                 sep: str =':') -> Dict:
    """
    Flatten a nested dictionary and create new keys by concatenation

    :param d: input dictionary (nested)
    :param parent_key: Prefix for keys of the flat dictionary
    :param sep: Separator when concatenating dictionary keys from different levels.
    """

    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class ColumnOverwriteException(Exception):
    """Raised when an existing column of a pandas dataframe is about to be overwritten"""
    pass


def stringify_list(cols):
    """
    Returns list with elements stringified
    """
    return [str(c) for c in cols]


def merge_dicts(d1: dict, d2: dict):
    """

    Merges two dicts

    :param d1: A dictionary
    :param d2: Another dictionary
    :return: Merged dicts
    """
    return dict(itertools.chain(d1.items(), d2.items()))


def get_context() -> mx.context:
    """

    Returns the a list of all available gpu contexts for a given machine.
    If no gpus are available, returns [mx.cpu()].
    Use it to automatically return MxNet contexts (uses max number of gpus or cpu)

    :return: List of mxnet contexts of a gpu or [mx.cpu()] if gpu not available

    """
    context_list = []
    for gpu_number in range(16):
        try:
            _ = mx.nd.array([1, 2, 3], ctx=mx.gpu(gpu_number))
            context_list.append(mx.gpu(gpu_number))
        except mx.MXNetError:
            pass

    if len(context_list) == 0:
        context_list.append(mx.cpu())

    return context_list


def random_split(data_frame: pd.DataFrame,
                 split_ratios: List[float] = None,
                 seed: int = 10) -> List[pd.DataFrame]:
    """

    Shuffles and splits a Data frame into partitions with specified percentages of data

    :param data_frame: a pandas DataFrame
    :param split_ratios: percentages of splits
    :param seed: seed of random number generator
    :return:
    """
    if split_ratios is None:
        split_ratios = [.8, .2]
    sections = np.array([int(r * len(data_frame)) for r in split_ratios]).cumsum()
    return np.split(data_frame.sample(frac=1, random_state=seed), sections)[:len(split_ratios)]


def rand_string(length: int = 16) -> str:
    """
    Utility function for generating a random alphanumeric string of specified length

    :param length: length of the generated string

    :return: random string
    """
    import random, string
    return ''.join([random.choice(string.ascii_letters + string.digits) for _ in range(length)])


@contextlib.contextmanager
def timing(marker):
    start_time = time.time()
    sys.stdout.flush()
    logger.info("\n========== start: %s" % marker)
    try:
        yield
    finally:
        logger.info("\n========== done (%s s) %s" % ((time.time() - start_time), marker))
        sys.stdout.flush()


class MeanSymbol(mx.metric.EvalMetric):
    """
    Metrics tracking the mean of a symbol, the index of the symbol must be passed as argument.
    """

    def __init__(self, name, symbol_index=0, output_names=None, label_names=None):
        super(MeanSymbol, self).__init__(name, output_names=output_names,
                                         label_names=label_names)
        self.symbol_index = symbol_index

    def update(self, _, preds):
        sym = preds[self.symbol_index]
        self.sum_metric += mx.ndarray.sum(sym).asscalar()
        self.num_inst += sym.size


class AccuracyMetric(mx.metric.EvalMetric):
    """
    Metrics tracking the accuracy, the index of the discrete label must be passed as argument.
    """

    def __init__(self, name, label_index=0):
        super(AccuracyMetric, self).__init__("{}-accuracy".format(name))
        self.label_index = label_index

    def update(self, labels, preds):
        chosen = preds[1 + self.label_index].asnumpy().argmax(axis=1)
        labels_values = labels[self.label_index].asnumpy().squeeze(axis=1)
        self.sum_metric += sum((chosen == labels_values) | (labels_values == 0.0))
        self.num_inst += preds[0].size


class LogMetricCallBack(object):
    """
    Tracked the metrics specified as arguments.
    Any mxnet metric whose name contains one of the argument will be tracked.
    """

    def __init__(self, tracked_metrics, patience=None):
        """
        :param tracked_metrics: metrics to be tracked
        :param patience: if not None then if the metrics does not improve during 'patience' number
        of steps, StopIteration is raised.
        """
        self.tracked_metrics = tracked_metrics
        self.metrics = {metric: [] for metric in tracked_metrics}
        self.patience = patience

    def __call__(self, param):
        if param.eval_metric is not None:
            name_value = param.eval_metric.get_name_value()
            for metric in self.tracked_metrics:
                for name, value in name_value:
                    if metric in name and not math.isnan(value):
                        self.metrics[metric].append(value)
                        if self.patience is not None:
                            self.check_regression()

    def check_regression(self):
        """
        If no improvement happens in "patience" number of steps then StopIteration exception is
        raised. This is an ugly mechanism but it is currently the only way to support this feature
        while using module api.
        """
        _, errors = next(iter(self.metrics.items()))

        def convert_nans(e):
            if math.isnan(e):
                logger.warning("Found nan in metric")
                return 0
            else:
                return e

        errors = [convert_nans(e) for e in errors]

        if self.patience < len(errors):
            # check that the metric has improved, e.g. that all recent metrics were worse
            metric_before_patience = errors[(-1 * self.patience) - 1]

            no_improvement = all(
                [errors[-i] >= metric_before_patience for i in range(0, self.patience)]
            )
            if no_improvement:
                logger.info("No improvement detected for {} epochs compared to {} last error " \
                            "obtained: {}, stopping here".format(self.patience,
                                                                 metric_before_patience,
                                                                 errors[-1]
                                                                 ))
                raise StopIteration


def normalize_dataframe(data_frame: pd.DataFrame):
    """

    Convenience function that normalizes strings of a DataFrame by converting to lower-case,
    stripping spaces, keeping only alphanumeric and space.

    :param data_frame:
    :return: normalized data_frame
    """
    return data_frame.apply(
        lambda x: x
            .astype(str)
            .str.lower()
            .str.replace(r'\s+', ' ')
            .str.strip()
            .str.replace('[^a-zA-Z0-9_ \r\n\t\f\v]', '')
    )


def softmax(x):
    """
    Compute softmax values for each sets of scores in x.
    """
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum()


def generate_df_string(word_length: int = 5,
                       vocab_size: int = 100,
                       num_labels: int = 2,
                       num_words: int = 5,
                       num_samples: int = 200,
                       label_column_name: str = "labels",
                       data_column_name: str = "sentences") -> pd.DataFrame:
    """
    Generates a dataframe with random strings in one column and random 'labels', which are
     substrings contained in the string column.

     Use this method for testing the imputer on string data

    :param word_length: length of the synthetic words
    :param vocab_size:  size of synthetic vocabulary
    :param num_labels:  number of labels / categories
    :param num_words:   number of words in each sentence
    :param n_samples:   number of samples in the data frame
    :param label_column_name: name of the label column
    :param data_column_name:  name of the data column
    :return:

    """
    vocab = [rand_string(word_length) for _ in range(vocab_size)]
    labels, words = vocab[:num_labels], vocab[num_labels:]

    def sentence_with_label(labels=labels, words=words):
        label = random.choice(labels)
        return " ".join(np.random.permutation([random.choice(words) for _ in range(num_words)] + [label])), label

    sentences, labels = zip(*[sentence_with_label(labels, words) for _ in range(num_samples)])

    return pd.DataFrame({data_column_name: sentences, label_column_name: labels})


def generate_df_numeric(num_samples: int = 100,
                        label_column_name: str = "f(x)",
                        data_column_name: str = "x") -> pd.DataFrame:
    """
    Generates a dataframe with random numbers between -pi and pi in one column and the square of those values in another

    :param num_samples:         number of samples to be generated
    :param label_column_name:   name of label column
    :param data_column_name:    name of data column
    :return:
    """
    numeric_data = np.random.uniform(-np.pi, np.pi, (num_samples,))
    return pd.DataFrame({
        data_column_name: numeric_data,
        label_column_name: numeric_data ** 2 + np.random.normal(0, .01, (num_samples,)),
    })


def random_cartesian_product(sets: List,
                             num: int = 10) -> List:
    """
    Return random samples from the cartesian product of all iterators in sets.
    Returns at most as many results as unique products exist.
    Does not require materialization of the full product but is still truly random,
    wich can't be achieved with itertools.

    Example usage:
    >>> random_cartesian_product([range(2**50), ['a', 'b']], num=2)
    >>> [[558002326003088, 'a'], [367785400774751, 'a']]

    :param sets: List of iteratbles
    :param num: Number of random samples to draw
    """

    # Determine cardinality of full cartisian product
    N = np.prod([len(y) for y in sets])

    # Draw random integers without replacement
    if N > 1e6:  # avoid materialising all integers if data is large
        idxs = []
        while len(idxs) < min(num, N):
            idx_candidate = np.random.randint(N)
            if idx_candidate not in idxs:
                idxs.append(idx_candidate)
    else:
        idxs = np.random.choice(range(N), size=min(num, N), replace=False)

    out = []
    for idx in idxs:
        out.append(sample_cartesian(sets, idx, N))

    return out


def sample_cartesian(sets: List,
                     idx: int,
                     n: int = None) -> List:
    """
    Draw a single sample from the cartesian product of all iterables in sets.
    Each row in the cartesian product has a unique index. This function returns
    the row with index idx without materialising any of the other rows.

    For a cartesian products of lists with length l1, l2, ... lm, taking the cartesian
    product can be thought of as traversing through all lists picking one element of each
    and repeating this until all possible combinations are exhausted. The number of combinations
    is N=l1*l2*...*lm. By taking the first element from every list that leads to a new combination,
    we can define a unique enumeration of all combinations.

    :param sets: List of iteratbles
    :param idx: Index of desired row in the cartersian product
    :param n: Number of rows in the cartesian product
    """

    if n is None:
        n = np.prod([len(y) for y in sets])

    out = []  # prepare list to append elements to.
    width = n  # width of the index set in which the desired row falls.
    for item_set in sets:
        width = width/len(item_set)  # map index set onto first item_set
        bucket = int(np.floor(idx/width))  # determine index of the first item_set
        out.append(item_set[bucket])
        idx = idx - bucket*width  # restrict index to next item_set in the hierarchy (could use modulo operator here.)

    assert width == 1  # at the end of this procedure, the leaf index set should have width 1.

    return out
